/*
 * Copyright 2012 Martin Skurla
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package org.crazyjavahacking.astvisualizer.netbeans;


import java.awt.Color;
import java.awt.Paint;
import java.util.IdentityHashMap;
import java.util.Map;

import org.abego.treelayout.Configuration;
import org.abego.treelayout.netbeans.AbegoTreeLayoutForNetbeans;
import org.crazyjavahacking.astvisualizer.api.CustomizedAST;
import org.crazyjavahacking.astvisualizer.api.model.EdgeModel;
import org.crazyjavahacking.astvisualizer.api.model.NodeModel;
import org.netbeans.api.visual.action.ActionFactory;
import org.netbeans.api.visual.action.HoverProvider;
import org.netbeans.api.visual.action.WidgetAction;
import org.netbeans.api.visual.anchor.Anchor;
import org.netbeans.api.visual.anchor.AnchorFactory;
import org.netbeans.api.visual.anchor.AnchorShape;
import org.netbeans.api.visual.border.BorderFactory;
import org.netbeans.api.visual.graph.GraphScene;
import org.netbeans.api.visual.layout.LayoutFactory;
import org.netbeans.api.visual.layout.SceneLayout;
import org.netbeans.api.visual.widget.ConnectionWidget;
import org.netbeans.api.visual.widget.LabelWidget;
import org.netbeans.api.visual.widget.LayerWidget;
import org.netbeans.api.visual.widget.Widget;


/**
 * AST graph scene for NetBeans Visual Library API.
 *
 * @author Martin Skurla (crazyjavahacking@gmail.com)
 */
public final class ASTGraphScene extends GraphScene<NodeModel, EdgeModel> {

    //-----------------------------------------------------------------------------------------------------------------
    // Constants.
    //-----------------------------------------------------------------------------------------------------------------
    private static final Paint DEFAULT_NODE_HOVER_PAINT = Color.RED;


    //-----------------------------------------------------------------------------------------------------------------
    // Members.
    //-----------------------------------------------------------------------------------------------------------------
    private final LayerWidget mainLayer;
    private final LayerWidget connectionLayer;
    private final Paint nodeHoverPaint;

    private final Map<NodeHoverListener, WidgetAction> nodeHoverListenerToWidgetActionMapper =
            new IdentityHashMap<NodeHoverListener, WidgetAction>();

    private NodeModel rootNodeModel = null;


    //-----------------------------------------------------------------------------------------------------------------
    // Constructors.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Creates ASTGraphScene from given customized AST. Default paint will be used for hovered node painting (Color.RED)
     * and abego Tree Layout will automatically be applied.
     *
     * @param customizedAST customized AST
     */
    public ASTGraphScene(CustomizedAST customizedAST) {
        this(customizedAST, DEFAULT_NODE_HOVER_PAINT);
    }

    /**
     * Creates ASTGraphScene from given customized AST and node hover paint. abego Tree Layout will automatically be
     * applied.
     *
     * @param customizedAST  customized AST
     * @param nodeHoverPaint node hover paint
     */
    public ASTGraphScene(CustomizedAST customizedAST, Paint nodeHoverPaint) {
        this.mainLayer       = new LayerWidget(this);
        this.connectionLayer = new LayerWidget(this);

        addChild(mainLayer);
        addChild(connectionLayer);

        this.nodeHoverPaint = nodeHoverPaint;

        fillWithCustomizedAST(customizedAST);

        applyTreeLayout();
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Public Methods.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * Applies abego Tree Layout with default configuration.
     *
     * @see ASTGraphSceneConfiguration
     */
    public void applyTreeLayout() {
        applyTreeLayout(new ASTGraphSceneConfiguration());
    }

    /**
     * Applies abego Tree Layout with given configuration.
     *
     * @param configuration configuration
     *
     * @since 1.0.2
     */
    public void applyTreeLayout(ASTGraphSceneConfiguration configuration) {
        AbegoTreeLayoutForNetbeans<NodeModel, EdgeModel> graphLayout =
            new AbegoTreeLayoutForNetbeans<NodeModel, EdgeModel>(rootNodeModel,
                                                                 configuration.getOriginX(),
                                                                 configuration.getOriginY(),
                                                                 new AbegoTreeLayoutConfigurationAdapter(configuration));

        SceneLayout sceneLayout = LayoutFactory.createSceneGraphLayout(this, graphLayout);

        sceneLayout.invokeLayoutImmediately();
    }

    /**
     * Registers NodeHoverListener.
     *
     * @param nodeHoverListener node hover listener
     */
    public void addNodeHoverListener(NodeHoverListener nodeHoverListener) {
        NodeHoverProviderAdapter nodeHoverProvider = new NodeHoverProviderAdapter(nodeHoverListener);

        WidgetAction hoverAction = ActionFactory.createHoverAction(nodeHoverProvider);

        for (Widget widget : mainLayer.getChildren()) {
            widget.getActions().addAction(hoverAction);
        }

        nodeHoverListenerToWidgetActionMapper.put(nodeHoverListener, hoverAction);
    }

    /**
     * Unregisters NodeHoverListener.
     *
     * @param nodeHoverListener node hover listener
     */
    public void removeNodeHoverListener(NodeHoverListener nodeHoverListener) {
        WidgetAction hoverAction = nodeHoverListenerToWidgetActionMapper.get(nodeHoverListener);

        for (Widget widget : mainLayer.getChildren()) {
            widget.getActions().removeAction(hoverAction);
        }

        nodeHoverListenerToWidgetActionMapper.remove(nodeHoverListener);
    }


    //-----------------------------------------------------------------------------------------------------------------
    // Private Helpers.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * {@inheritDoc}
     */
    @Override
    protected Widget attachNodeWidget(NodeModel nodeModel) {
        if (rootNodeModel == null) {
            rootNodeModel = nodeModel;
        }

        LabelWidget widget = new LabelWidget(this, nodeModel.getTreeKind());
        widget.setToolTipText(nodeModel.getTextualRepresentation());
        widget.setBorder(BorderFactory.createLineBorder());

        mainLayer.addChild(widget);

        return widget;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Widget attachEdgeWidget(EdgeModel edge) {
        ConnectionWidget connectionWidget = new ConnectionWidget(this);
        connectionWidget.setTargetAnchorShape(AnchorShape.TRIANGLE_HOLLOW);

        LabelWidget labelWidget = new LabelWidget(this, edge.getTransitionName());
        labelWidget.setOpaque(true);

        connectionWidget.addChild(labelWidget);
        connectionWidget.setConstraint(labelWidget, LayoutFactory.ConnectionWidgetLayoutAlignment.CENTER, 0.5F);

        connectionLayer.addChild(connectionWidget);

        return connectionWidget;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void attachEdgeSourceAnchor(EdgeModel edge, NodeModel oldNodeModel, NodeModel nodeModel) {
        ConnectionWidget connectionWidget = (ConnectionWidget) findWidget(edge);

        Widget sourceWidget = findWidget(nodeModel);

        Anchor sourceAnchor = AnchorFactory.createRectangularAnchor(sourceWidget);

        connectionWidget.setSourceAnchor(sourceAnchor);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void attachEdgeTargetAnchor(EdgeModel edge, NodeModel oldNodeModel, NodeModel nodeModel) {
        ConnectionWidget connectionWidget = (ConnectionWidget) findWidget(edge);

        Widget targetWodget = findWidget(nodeModel);

        Anchor targetAnchor = AnchorFactory.createRectangularAnchor(targetWodget);

        connectionWidget.setTargetAnchor(targetAnchor);
    }

    /**
     * Inserts all nodes and edges from given customized AST.
     *
     * @param customizedAST  customized AST
     */
    private void fillWithCustomizedAST(CustomizedAST customizedAST) {
        for (NodeModel nodeModel : customizedAST) {
            this.addNode(nodeModel);

            EdgeModel incomingEdge = customizedAST.getIncomingEdgeOfNode(nodeModel);

            if (incomingEdge != EdgeModel.NO_INCOMING_EDGE) {
                NodeModel edgeSource = customizedAST.getSourceNodeOfEdge(incomingEdge);
                NodeModel edgeTarget = customizedAST.getTargetNodeOfEdge(incomingEdge);

                this.addEdge(incomingEdge);
                this.setEdgeSource(incomingEdge, edgeSource);
                this.setEdgeTarget(incomingEdge, edgeTarget);
            }
        }
    }

    //-----------------------------------------------------------------------------------------------------------------
    // Inner classes.
    //-----------------------------------------------------------------------------------------------------------------
    /**
     * NetBeans Visual Library HoverProvider adapter.
     */
    private final class NodeHoverProviderAdapter implements HoverProvider {

        //-------------------------------------------------------------------------------------------------------------
        // Members.
        //-------------------------------------------------------------------------------------------------------------
        private final NodeHoverListener nodeHoverListener;

        private Widget previousWidget = null;
        private Paint previousWidgetBackgroundPaint = null;


        //-------------------------------------------------------------------------------------------------------------
        // Constructors.
        //-------------------------------------------------------------------------------------------------------------
        private NodeHoverProviderAdapter(NodeHoverListener nodeHoverListener) {
            this.nodeHoverListener = nodeHoverListener;
        }


        //-------------------------------------------------------------------------------------------------------------
        // Public Methods.
        //-------------------------------------------------------------------------------------------------------------
        /**
         * {@inheritDoc}
         */
        @Override
        public void widgetHovered(Widget widget) {
            if (previousWidget != null) {
                previousWidget.setBackground(previousWidgetBackgroundPaint);
            }

            previousWidget = widget;
            previousWidgetBackgroundPaint = widget.getBackground();

            widget.setOpaque(true);
            widget.setBackground(ASTGraphScene.this.nodeHoverPaint);

            NodeModel selectedNodeModel = (NodeModel) findObject(widget);

            nodeHoverListener.nodeHovered(selectedNodeModel);
        }
    }

    /**
     * abego Tree Layout configuration adapter.
     */
    private static final class AbegoTreeLayoutConfigurationAdapter implements Configuration<NodeModel> {
        private final ASTGraphSceneConfiguration graphSceneConfiguration;

        private AbegoTreeLayoutConfigurationAdapter(ASTGraphSceneConfiguration graphSceneConfiguration) {
            this.graphSceneConfiguration = graphSceneConfiguration;
        }

        @Override
        public Location getRootLocation() {
            switch (graphSceneConfiguration.getRootNodeLocation()) {
                case BOTTOM:
                    return Location.Bottom;
                case LEFT:
                    return Location.Left;
                case RIGHT:
                    return Location.Right;
                case TOP:
                    return Location.Top;
                default:
                    throw new IllegalStateException(
                            "Given RootNodeLocation " + graphSceneConfiguration.getRootNodeLocation() + " not supported !");
            }
        }

        @Override
        public AlignmentInLevel getAlignmentInLevel() {
            return AlignmentInLevel.Center;
        }

        @Override
        public double getGapBetweenLevels(int nextLevel) {
            return graphSceneConfiguration.getGapBetweenLevels();
        }

        @Override
        public double getGapBetweenNodes(NodeModel node1, NodeModel node2) {
            return graphSceneConfiguration.getGapBetweenNodes();
        }

    }
}
